using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Serialization;
using System.Xml.XPath;
using System.Xml.Xsl;
using EVEMon.Common;
using EVEMon.Common.Schedule;
using EVEMon.SkillPlanner;
using System.Web;

namespace EVEMon
{
    public partial class CharacterMonitor : UserControl
    {
        private Settings m_settings;
        private CharLoginInfo m_cli;
        private SerializableCharacterInfo m_sci;
        private CharFileInfo m_cfi;
        private EveSession m_session;
        private int m_charId;
        private FileSystemWatcher m_characterFileWatch = null;
        private Dictionary<SkillGroup, bool> m_groupCollapsed = new Dictionary<SkillGroup, bool>();
        private DateTime m_nextScheduledUpdateAt = DateTime.MinValue;
        private string m_skillTrainingName;
        private DateTime m_estimatedCompletion;
        private CharacterInfo m_grandCharacterInfo;
        private int m_lastTickSPPaint = 0;
        private bool m_updatingPortrait = false;
        private bool m_currentlyVisible = false;

        //This is data that would be shown ON the LCD display, not data ABOUT the LCD display
        public EventHandler LCDDataChanged;
        private string m_lcdNote = String.Empty;
        private TimeSpan m_lcdTimeSpan = TimeSpan.Zero;

        //file formats for saving your character
        private enum SaveFormat
        {
            None = 0,
            Text = 1,
            Html = 2,
            Xml = 3,
            Png = 4
        }

        private CharacterMonitor()
        {
            InitializeComponent();
            SetStyle(ControlStyles.SupportsTransparentBackColor, true);

        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CharacterMonitor"/> class.  For eve-o based characters
        /// </summary>
        /// <param name="s">The EVEMon settings.</param>
        /// <param name="cli">The character login info.</param>
        public CharacterMonitor(Settings s, CharLoginInfo cli)
            : this()
        {
            m_settings = s;
            m_cli = cli;
            m_sci = null;
            m_cfi = null;
        }


        /// <summary>
        /// Initializes a new instance of the <see cref="CharacterMonitor"/> class.  For file based characters
        /// </summary>
        /// <param name="s">The EVEMon settings.</param>
        /// <param name="cfi">The character file info.</param>
        /// <param name="sci">The serializable character, loaded from the file.</param>
        public CharacterMonitor(Settings s, CharFileInfo cfi, SerializableCharacterInfo sci)
            : this()
        {
            m_settings = s;
            m_cli = null;
            m_sci = sci;
            m_cfi = cfi;
        }

        private string m_charName
        {
            get
            {
                if (m_grandCharacterInfo != null && m_grandCharacterInfo.Name != null)
                {
                    return m_grandCharacterInfo.Name;
                }
                else
                {
                    return "?";
                }
            }
        }

        /// <summary>
        /// Starts the monitor.  Sets up the GrandCharacterInfo, attempts to get the image, starts all the timers and, if needed,
        /// the FileSystemWatcher
        /// <remarks>This method can probably be broken up quite a bit, for simplicity's sake</remarks>
        /// </summary>
        public void Start()
        {
            m_session = null;
            m_charId = -1;

            if (m_cli != null)
            {
                m_grandCharacterInfo = new CharacterInfo(m_charId, m_cli.CharacterName);
            }
            else
            {
                m_charId = m_sci.CharacterId;
                m_grandCharacterInfo = new CharacterInfo(m_sci.CharacterId, m_sci.Name);
                if (m_charId > 0)
                {
                    GetCharacterImage();
                }
                if (m_cfi.MonitorFile)
                {
                    m_characterFileWatch = new FileSystemWatcher(
                        Path.GetDirectoryName(m_cfi.Filename), Path.GetFileName(m_cfi.Filename));
                    m_characterFileWatch.Created += new FileSystemEventHandler(CharacterFileUpdatedCallback);
                    m_characterFileWatch.Changed += new FileSystemEventHandler(CharacterFileUpdatedCallback);
                    m_characterFileWatch.EnableRaisingEvents = true;
                }
            }
            m_grandCharacterInfo.BioInfoChanged += new EventHandler(BioInfoChangedCallback);
            m_grandCharacterInfo.BalanceChanged += new EventHandler(BalanceChangedCallback);
            m_grandCharacterInfo.AttributeChanged += new EventHandler(AttributeChangedCallback);
            m_grandCharacterInfo.SkillChanged += new SkillChangedHandler(CharacterSkillChangedCallback);
            m_grandCharacterInfo.TrainingSkillChanged += new EventHandler(TrainingSkillChangedCallback);

            if (m_cli != null)
            {
                SerializableCharacterInfo sci = m_settings.GetCharacterInfo(m_cli.CharacterName);
                if (sci != null)
                {
                    m_grandCharacterInfo.AssignFromSerializableCharacterInfo(sci);
                }
                if (m_settings.DisableXMLAutoUpdate == false)
                {
                    tmrUpdateCharacter.Interval = 10;
                    tmrUpdateCharacter.Enabled = true;
                }
                else
                {
                    tmrUpdateCharacter.Enabled = false;
                    pbThrobber.Visible = false;
                }
            }
            else
            {
                tmrUpdateCharacter.Enabled = false;
                pbThrobber.Visible = false;
                m_grandCharacterInfo.AssignFromSerializableCharacterInfo(m_sci);
                m_sci = null;
            }

            foreach (SkillGroup gsg in m_grandCharacterInfo.SkillGroups.Values)
            {
                foreach (Pair<string, string> grp in m_settings.CollapsedGroups)
                {
                    if ((grp.A == m_grandCharacterInfo.Name) && (grp.B == gsg.Name))
                    {
                        m_groupCollapsed.Add(gsg, true);
                        foreach (Skill gs in gsg)
                        {
                            if (gs.Known)
                            {
                                lbSkills.Items.RemoveAt(lbSkills.Items.IndexOf(gs));
                            }
                        }
                        gsg.isCollapsed = true;
                    }
                }
            }

            tmrTick.Enabled = true;

            m_settings.WorksafeChanged += new EventHandler<EventArgs>(WorksafeChangedCallback);
            WorksafeChangedCallback(null, null);

            m_settings.UseLogitechG15DisplayChanged += new EventHandler<EventArgs>(UpdateLcdDisplayCallback);

            if (m_settings != null)
            {
                //set up individual character settings
                tsbIneveSync.Checked = m_settings.GetCharacterSettings(m_charName).IneveSync;
            }
        }

        private void ReloadFromFile()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(ReloadFromFile));
                return;
            }

            SerializableCharacterInfo sci = SerializableCharacterInfo.CreateFromFile(m_cfi.Filename);
            if (sci != null)
            {
                m_grandCharacterInfo.AssignFromSerializableCharacterInfo(sci);
            }
        }

        /// <summary>
        /// Stop all timers, unhook all events, lie down and die.
        /// </summary>
        public void Stop()
        {
            if (m_session != null)
            {
                m_session = null;
            }
            tmrTick.Enabled = false;
            tmrUpdateCharacter.Enabled = false;
            if (m_characterFileWatch != null)
            {
                m_characterFileWatch.EnableRaisingEvents = false;
            }

            m_grandCharacterInfo.BioInfoChanged -= new EventHandler(BioInfoChangedCallback);
            m_grandCharacterInfo.BalanceChanged -= new EventHandler(BalanceChangedCallback);
            m_grandCharacterInfo.AttributeChanged -= new EventHandler(AttributeChangedCallback);
            m_grandCharacterInfo.SkillChanged -= new SkillChangedHandler(CharacterSkillChangedCallback);
            m_grandCharacterInfo.TrainingSkillChanged -= new EventHandler(TrainingSkillChangedCallback);

            m_settings.WorksafeChanged -= new EventHandler<EventArgs>(WorksafeChangedCallback);
            m_settings.UseLogitechG15DisplayChanged -= new EventHandler<EventArgs>(UpdateLcdDisplayCallback);
        }

        /// <summary>
        /// Gets or sets a value indicating whether the panel is currently visible
        /// </summary>
        /// <value><c>true</c> if visible; otherwise, <c>false</c>.</value>
        public bool CurrentlyVisible
        {
            get
            {
                return m_currentlyVisible;
            }
            set
            {
                bool old_value = m_currentlyVisible;
                m_currentlyVisible = value;
                UpdateLcdisplay();
                if (m_currentlyVisible != old_value && m_currentlyVisible)
                {
                    UpdateSkillHeaderStats();
                }
            }
        }

        /// <summary>
        /// Enables the three main buttons on the form.
        /// </summary>
        private void EnableButtons()
        {
            btnSave.Enabled = true;
            btnPlan.Enabled = true;
            btnMoreOptions.Enabled = true;
        }

        /// <summary>
        /// Updates skill summary info (the block just below the portrait).
        /// </summary>
        private void UpdateSkillHeaderStats()
        {
            lblSkillHeader.Text = String.Format("{0} Known Skills\n{1} Total SP\n{2} Skills at Level V",
                                                m_grandCharacterInfo.KnownSkillCount,
                                                m_grandCharacterInfo.SkillPointTotal.ToString("#,##0"),
                                                m_grandCharacterInfo.MaxedSkillCount);
        }

        /// <summary>
        /// Sets the specified attribute label.
        /// </summary>
        /// <param name="lblAttrib">The LBL attrib.</param>
        /// <param name="eveAttribute">The eve attribute.</param>
        private void SetAttributeLabel(Label lblAttrib, EveAttribute eveAttribute)
        {
            StringBuilder sb = new StringBuilder();

            sb.Append(eveAttribute.ToString());

            sb.Append(": ");
            sb.Append(m_grandCharacterInfo.GetEffectiveAttribute(eveAttribute).ToString("0.00"));
            lblAttrib.Text = sb.ToString();
        }

        /// <summary>
        /// Updates the cached copy.
        /// </summary>
        /// <remarks>This makes sure the settings file only has you down with one "last skill trained", then saves the settings file</remarks>
        private void UpdateCachedCopy()
        {
            SerializableCharacterInfo sci = m_grandCharacterInfo.ExportSerializableCharacterInfo();
            m_settings.SetCharacterCache(sci);
            if (m_grandCharacterInfo.OldTrainingSkill != null && m_grandCharacterInfo.OldTrainingSkill.old_SkillName != null)
            {
                int toremove;
                bool add = true;
                do
                {
                    toremove = -1;
                    foreach (Pair<string, OldSkillinfo> x in m_settings.OldSkillLearnt)
                    {
                        if (x.A == m_charName)
                        {
                            if (x.B.old_SkillName == m_grandCharacterInfo.OldTrainingSkill.old_SkillName && x.B.old_TrainingToLevel == m_grandCharacterInfo.OldTrainingSkill.old_TrainingToLevel)
                            {
                                add = false;
                                x.B.old_skill_completed = m_grandCharacterInfo.OldTrainingSkill.old_skill_completed;
                                x.B.old_estimated_completion = m_grandCharacterInfo.OldTrainingSkill.old_estimated_completion;
                                if (!x.B.old_skill_completed)
                                {
                                    toremove = m_settings.OldSkillLearnt.IndexOf(x);
                                }
                            }
                            else
                            {
                                toremove = m_settings.OldSkillLearnt.IndexOf(x);
                            }
                        }
                    }
                    if (toremove != -1)
                        m_settings.OldSkillLearnt.RemoveAt(toremove);
                } while (toremove != -1);
                if (add && m_grandCharacterInfo.OldTrainingSkill.old_skill_completed)
                    m_settings.OldSkillLearnt.Add(new Pair<string, OldSkillinfo>(m_charName, m_grandCharacterInfo.OldTrainingSkill));
            }
            m_settings.Save();
        }

        private void DownloadCharacter()
        {
#if DEBUG_SINGLETHREAD
            GetCharIdAndUpdateInternal(null);
#else
            ThreadPool.QueueUserWorkItem(new WaitCallback(GetCharIdAndUpdateInternal));
#endif
        }

        /// <summary>
        /// Actually download and update the character
        /// </summary>
        /// <param name="state">Does nothing, provided for compliance with WaitCallback</param>
        private void GetCharIdAndUpdateInternal(object state)
        {
            //get the character ID
            int charId = -1;
            try
            {
                m_session = EveSession.GetSession(m_cli.Username, m_cli.Password);
                charId = m_session.GetCharacterId(m_cli.CharacterName);
            }
            catch (Exception e)
            {
                ExceptionHandler.LogException(e, false);
            }
            if (charId < 0)
            {
                this.Invoke(new MethodInvoker(CharacterDownloadFailedCallback));
                return;
            }

            m_grandCharacterInfo.CharacterId = charId;
            GetCharacterImage();

            //actually perform the update.  This is then happening asynchronously, is this a problem?
            UpdateGrandCharacterInfo();            

            //loads the details from the settings file as to what was actually training the last time you had EVEMon running
            foreach (Pair<string, OldSkillinfo> x in m_settings.OldSkillLearnt)
            {
                if (x.A == m_charName)
                {
                    m_grandCharacterInfo.OldTrainingSkill = new OldSkillinfo(x.B.old_SkillName, x.B.old_TrainingToLevel, x.B.old_skill_completed, x.B.old_estimated_completion);
                }
            }
        }

        private void UpdateGrandCharacterInfo()
        {
            m_session.UpdateGrandCharacterInfoAsync(m_grandCharacterInfo, Program.MainWindow,
                                                    new UpdateGrandCharacterInfoCallback(GrandCharacterUpdatedCallback));

        }

        /// <summary>
        /// Updates the throbber timer when finished updating
        /// </summary>
        /// <param name="s">The eve session - not used.</param>
        /// <param name="timeLeftInCache">The time left in cache - that is, the time before the next update to the sheet on eve-o.</param>
        private void GrandCharacterUpdatedCallback(EveSession s, int timeLeftInCache)
        {
            this.Invoke(new MethodInvoker(delegate
                                                                                              {
                                                                                                  //m_lastUpdate = DateTime.Now;
                                                                                                  m_nextScheduledUpdateAt = DateTime.Now + TimeSpan.FromMilliseconds(timeLeftInCache);
                                                                                                  ttToolTip.SetToolTip(pbThrobber, "Click to update now.");
                                                                                                  ttToolTip.IsBalloon = true;
                                                                                                  //timeLeftInCache == 0 is the same as timeLeftInCache == 60 minutes.
                                                                                                  tmrUpdateCharacter.Interval = timeLeftInCache == 0 ? 3600000 : timeLeftInCache;
                                                                                                  tmrUpdateCharacter.Enabled = true;
                                                                                                  StopThrobber();
                                                                                              }));
        }

        /// <summary>
        /// Updates the LCD Display if G15 support is on
        /// </summary>
        private void UpdateLcdisplay()
        {
            if (m_settings.UseLogitechG15Display && Program.LCD != null)
            {
                if (Program.LCD.CharacterName == null || Program.LCD.CharacterName == "")
                {
                    Program.LCD.CharacterName = m_charName;
                    Program.LCD.newchar = m_charName;
                }
                if (Program.LCD.newchar == m_charName)
                {
                    Skill s = m_grandCharacterInfo.CurrentlyTrainingSkill;
                    if (s != null)
                    {
                       Program.LCD.curperc = s.GetPercentDone();
                    }
                    else
                    {
                        Program.LCD.curperc = 1;
                    }
                    Program.LCD.CharacterName = m_charName; //null ok
                    Program.LCD.CurrentSkillTrainingText = m_skillTrainingName; //null ok
                    Program.LCD.TimeToComplete = m_lcdTimeSpan;
                }
                long tmp = Program.LCD.leasttime.Ticks - m_lcdTimeSpan.Ticks;
                if (tmp > 0)
                {
                    if (m_skillTrainingName != "" && m_skillTrainingName != null)
                    {
                        Program.LCD.leasttime = m_lcdTimeSpan;
                        Program.LCD.leastchar = m_charName;
                    }
                }
                if (Program.LCD.charlist == null)
                {
                    List<CharLoginInfo> mlist = m_settings.CharacterList;
                    string[] temp = new string[mlist.Count];
                    int i = 0;
                    foreach (CharLoginInfo sci in mlist)
                    {
                        temp[i] = sci.CharacterName;
                        i++;
                    }
                    Program.LCD.charlist = temp;
                }
                if (Program.LCD.refreshchar != null && Program.LCD.refreshchar == m_charName)
                {
                    Program.LCD.refreshchar = null;
                    tmrUpdate_Tick(null, null);
                    UpdateThrobberLabel();
                }
                if (Program.LCD.cycleini == false)
                {
                    Program.LCD.cycle = m_settings.G15ACycle;
                    Program.LCD.cycleint = m_settings.G15ACycleint;
                    Program.LCD.cycleini = true;
                }
                else
                {
                    m_settings.G15ACycle = Program.LCD.cycle;
                }
            }
        }

        /// <summary>
        /// Calculate the skill text and time remaining for LCD Display
        /// </summary>
        private void CalculateLcdData()
        {
            DateTime now = DateTime.Now;
            if (m_estimatedCompletion != DateTime.MaxValue)
            {
                if (m_estimatedCompletion > now)
                {
                    SetLcdData(m_charName + ": " +
                             TimeSpanDescriptiveShort(m_estimatedCompletion),
                                 m_estimatedCompletion - now);
                }
                else
                {
                    SetLcdData(m_charName + ": Done", TimeSpan.Zero);
                }
            }
            else
            {
                SetLcdData(String.Empty, TimeSpan.Zero);
            }
        }

        /// <summary>
        /// Update labels in the bottom panel for time remaining
        /// </summary>
        private void UpdateTimeRemainingLabels()
        {
            DateTime now = DateTime.Now;
            if (m_estimatedCompletion != DateTime.MaxValue)
            {
                lblTrainingRemain.Text = TimeSpanDescriptive(m_estimatedCompletion);
                if (m_estimatedCompletion > now)
                {
                    lblTrainingEst.Text = m_estimatedCompletion.ToString();
                }
                else
                {
                    lblTrainingEst.Text = String.Empty;
                }
            }
            else
            {
                lblTrainingRemain.Text = String.Empty;
                lblTrainingEst.Text = String.Empty;
            }
        }

        /// <summary>
        /// Sets the note and the timespan used by the LCDdisplay functions
        /// </summary>
        /// <param name="newShortText">The new short text.</param>
        /// <param name="timeSpan">The time span.</param>
        private void SetLcdData(string newShortText, TimeSpan timeSpan)
        {
            bool fireEvent = false;
            if (newShortText != m_lcdNote || timeSpan != m_lcdTimeSpan)
            {
                fireEvent = true;
            }
            m_lcdNote = newShortText;
            m_lcdTimeSpan = timeSpan;
            if (fireEvent && LCDDataChanged != null)
            {
                LCDDataChanged(this, new EventArgs());
            }
            UpdateLcdisplay();
        }

        /// <summary>
        /// Returns a short timespan string between now and the DateTime given... which is in the future.
        /// If someone comes up with a clean way to whack this with string.format, go for it... as it is, it's a custom format, and not one
        /// that we can throw out, so this needs to stay.  It's not SO bad.
        /// Changes to this should also touch TimeSpanDescriptive.
        /// </summary>
        /// <param name="t">The time (in the future) for which you want a span.</param>
        /// <returns>Format:XdXh0Xm0Xs</returns>
        public static string TimeSpanDescriptiveShort(DateTime t)
        {
            StringBuilder sb = new StringBuilder();
            if (t > DateTime.Now)
            {
                TimeSpan ts = t - DateTime.Now;
                if (ts.Days > 0)
                {
                    sb.Append(ts.Days.ToString());
                    sb.Append("d");
                }
                ts -= TimeSpan.FromDays(ts.Days);
                if (ts.Hours > 0)
                {
                    sb.Append(ts.Hours.ToString());
                    sb.Append("h");
                }
                ts -= TimeSpan.FromHours(ts.Hours);
                if (ts.Minutes > 0)
                {
                    sb.Append(ts.Minutes.ToString());
                    sb.Append("m");
                }
                ts -= TimeSpan.FromMinutes(ts.Minutes);
                if (ts.Seconds > 0)
                {
                    sb.Append(ts.Seconds.ToString());
                    sb.Append("s");
                }
                return sb.ToString();
            }
            else
            {
                return "Done";
            }
        }

        /// <summary>
        /// Returns a long format timespan string between now and the DateTime given, which should be in the future.
        /// If someone comes up with a clean way to whack this with string.format, go for it... as it is, it's a custom format, and not one
        /// that we can throw out, so this needs to stay.  It's not SO bad.
        /// Changes to this should also touch TimeSpanDescriptiveShort.
        /// </summary>
        /// <param name="t">The time (in the future) for which you want a span</param>
        /// <returns>Format: X day(x), X hour(s), X minute(s), X second(s)</returns>
        private string TimeSpanDescriptive(DateTime t)
        {
            StringBuilder sb = new StringBuilder();
            if (t > DateTime.Now)
            {
                TimeSpan ts = t - DateTime.Now;
                if (ts.Days > 0)
                {
                    sb.Append(ts.Days.ToString());
                    sb.Append(" day");
                    if (ts.Days > 1)
                    {
                        sb.Append("s");
                    }
                }
                ts -= TimeSpan.FromDays(ts.Days);
                if (ts.Hours > 0)
                {
                    if (sb.Length > 0)
                    {
                        sb.Append(", ");
                    }
                    sb.Append(ts.Hours.ToString());
                    sb.Append(" hour");
                    if (ts.Hours > 1)
                    {
                        sb.Append("s");
                    }
                }
                ts -= TimeSpan.FromHours(ts.Hours);
                if (ts.Minutes > 0)
                {
                    if (sb.Length > 0)
                    {
                        sb.Append(", ");
                    }
                    sb.Append(ts.Minutes.ToString());
                    sb.Append(" minute");
                    if (ts.Minutes > 1)
                    {
                        sb.Append("s");
                    }
                }
                ts -= TimeSpan.FromMinutes(ts.Minutes);
                if (ts.Seconds > 0)
                {
                    if (sb.Length > 0)
                    {
                        sb.Append(", ");
                    }
                    sb.Append(ts.Seconds.ToString());
                    sb.Append(" second");
                    if (ts.Seconds > 1)
                    {
                        sb.Append("s");
                    }
                }
                return sb.ToString();
            }
            else
            {
                return "Completed.";
            }
        }

        /// <summary>
        /// Updates label above the throbber for the next update
        /// </summary>
        private void UpdateThrobberLabel()
        {
            if (m_throbberRunning)
            {
                lblUpdateTimer.Visible = false;
                return;
            }

            TimeSpan ts = m_nextScheduledUpdateAt - DateTime.Now;
            if (ts < TimeSpan.Zero || ts > TimeSpan.FromMinutes(60))
            {
                lblUpdateTimer.Visible = false;
            }
            else
            {
                lblUpdateTimer.Text = String.Format("{0:d2}:{1:d2}", ts.Minutes, ts.Seconds);
                lblUpdateTimer.Visible = true;
            }
        }

        /// <summary>
        /// Gets the character image.  First attempt to get it from cache, then attempt to get it from eve-o.  If you can't get it either place, fail quietly.
        /// </summary>
        private void GetCharacterImage()
        {
            if (pbCharImage.Image == null)
            {
                string cacheFileName = Settings.EveMonData + "\\cache\\" + this.GrandCharacterInfo.CharacterId.ToString() + ".png";
                if (File.Exists(cacheFileName))
                {
                    pbCharImage.Image = PortraitFromCache(cacheFileName);
                    m_updatingPortrait = false;
                }
                if (pbCharImage.Image == null)
                {
                    UpdateCharacterImageRemote();
                }
            }
        }
        private void UpdateCharacterImageRemote()
        {
            pbCharImage.Image = null;
            EveSession.GetCharacterImageAsync(this.GrandCharacterInfo.CharacterId,
                                              new GetImageCallback(GotCharacterImage));
        }

        /// <summary>
        /// Scan the specified EVE Folder for character portraits and, if possible, copy one of them to EVEMon's cache for use, then update the portrait display.
        /// </summary>
        private void UpdateCharacterImageLocal()
        {
            pbCharImage.Image = null;
            Image i = null;

            // test to see if an EVE Folder has been selected
            if (this.GrandCharacterInfo.EVEFolder == "")
            {
                // if the folder has not been selected, ask the user to select it
                RequestEVEFolder();
            }

            try
            {
                // generate paths and file patterns required for 
                string eveCacheFolder = this.GrandCharacterInfo.EVEFolder + "\\cache\\Pictures\\Portraits\\";
                string cacheFileName = Settings.EveMonData + "\\cache\\" + this.GrandCharacterInfo.CharacterId.ToString() + ".png";
                int charIDLength = this.GrandCharacterInfo.CharacterId.ToString().Length;

                // create a pattern that matches anything "<characterId>*.png"
                string filePattern = this.GrandCharacterInfo.CharacterId.ToString() + "*.png";

                // enumerate files in the EVE cache directory
                DirectoryInfo di = new DirectoryInfo(eveCacheFolder);
                FileInfo[] filesInEveCache = di.GetFiles(filePattern);

                if (filesInEveCache.Length > 0)
                {
                    // add each file to a sorted list
                    SortedList<int, string> sl = new SortedList<int, string>();

                    foreach (FileInfo file in filesInEveCache)
                    {
                        int sizeLength = (file.Name.Length - (file.Extension.Length + 1)) - charIDLength;
                        int imageSize = int.Parse(file.Name.Substring(charIDLength + 1, sizeLength));
                        sl.Add(imageSize, file.FullName);
                    }

                    // pick the largest image
                    string eveCacheFileName = sl.Values[sl.Count - 1];

                    FileStream fs1 = new FileStream(eveCacheFileName, FileMode.Open);
                    FileStream fs2 = new FileStream(cacheFileName, FileMode.Create);
                    i = Image.FromStream(fs1, true);
                    i.Save(fs2, ImageFormat.Png);
                    fs1.Close();
                    fs1.Dispose();
                    fs2.Close();
                    fs2.Dispose();
                }
                else
                {
                    String message;

                    message = "No portraits for your character were found in the folder you selected." + Environment.NewLine + Environment.NewLine;
                    message += "Ensure that you have checked the following" + Environment.NewLine;
                    message += " - You have selected a valid EVE Folder (i.e. C:\\Program Files\\CCP\\EVE\\)" + Environment.NewLine;
                    message += " - You have logged into EVE using this folder";

                    MessageBox.Show(message, "Portrait Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }
            }
            catch (Exception e)
            {
                ExceptionHandler.LogException(e, false);
            }

            pbCharImage.Image = i;
            m_updatingPortrait = false;
        }

        /// <summary>
        /// Pops up a window for the user to select their EVE folder.
        /// </summary>
        private void RequestEVEFolder()
        {
            using (EVEFolderWindow f = new EVEFolderWindow())
            {
                f.EVEFolder = m_grandCharacterInfo.EVEFolder;
                f.ShowDialog();
                if (f.DialogResult == DialogResult.OK)
                {
                    m_grandCharacterInfo.EVEFolder = f.EVEFolder;
                    UpdateCachedCopy();
                }
            }
        }

        /// <summary>
        /// Toggles the expansion or collapsing of a single group
        /// </summary>
        /// <param name="gsg">The group to expand or collapse.</param>
        private void ToggleGroupExpandCollapse(SkillGroup gsg)
        {
            bool toCollapse;
            if (gsg.isCollapsed)
            {
                toCollapse = false;
            }
            else
            {
                toCollapse = true;
            }
            m_groupCollapsed[gsg] = toCollapse;
            gsg.isCollapsed = m_groupCollapsed[gsg];
            if (toCollapse)
            {
                // Remove the skills in the group from the list
                foreach (Skill gs in gsg)
                {
                    if (gs.Known)
                    {
                        lbSkills.Items.RemoveAt(lbSkills.Items.IndexOf(gs));
                    }
                }

                Pair<string, string> grp = new Pair<string, string>(m_grandCharacterInfo.Name, gsg.Name);
                m_settings.CollapsedGroups.Add(grp);
            }
            else
            {
                List<Skill> skillList = new List<Skill>();
                foreach (Skill gs in gsg)
                {
                    skillList.Add(gs);
                }
                SkillChangedEventArgs args = new SkillChangedEventArgs(skillList.ToArray());
                CharacterSkillChangedCallback(this, args);
                foreach (Pair<string, string> grp in m_settings.CollapsedGroups)
                {
                    if ((grp.A == m_grandCharacterInfo.Name) && (grp.B == gsg.Name))
                    {
                        m_settings.CollapsedGroups.Remove(grp);
                        break;
                    }
                }
                //void m_grandCharacterInfo_SkillChanged(object sender, SkillChangedEventArgs e)
            }
            lbSkills.Invalidate(lbSkills.GetItemRectangle(lbSkills.Items.IndexOf(gsg)));
        }

        /// <summary>
        /// Handles the LinkClicked event of the llToggleAll control.  Toggles all the skill groups to collapse or open.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.LinkLabelLinkClickedEventArgs"/> instance containing the event data.</param>
        private void llToggleAll_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            List<SkillGroup> toggles = new List<SkillGroup>();
            bool? setCollapsed = null;
            foreach (object o in lbSkills.Items)
            {
                if (o is SkillGroup)
                {
                    SkillGroup gsg = (SkillGroup)o;
                    bool isCollapsed = (m_groupCollapsed.ContainsKey(gsg) && m_groupCollapsed[gsg]);
                    if (setCollapsed == null)
                    {
                        setCollapsed = !isCollapsed;
                    }
                    if (isCollapsed != setCollapsed)
                    {
                        toggles.Add(gsg);
                    }
                }
            }
            lbSkills.BeginUpdate();
            try
            {
                foreach (SkillGroup toggroup in toggles)
                {
                    ToggleGroupExpandCollapse(toggroup);
                }
            }
            finally
            {
                lbSkills.EndUpdate();
            }
        }

        /// <summary>
        /// Forces the LCD data to be recalculated.
        /// </summary>
        public void ForceLcdDataUpdate()
        {
            CalculateLcdData();
        }

        #region External Callbacks

        /// <summary>
        /// Updates the LCD display.  Used to allow external agents (ie the settings object) to force this monitor to update its LCD display data in the event that
        /// something that would influence LCD data changes.
        /// </summary>
        /// <param name="sender">Sender</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void UpdateLcdDisplayCallback(object sender, EventArgs e)
        {
            UpdateLcdisplay();
        }

        /// <summary>
        /// Called when the character file has been updated on the disc.  Used only by xml-backed characters
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.IO.FileSystemEventArgs"/> instance containing the event data.</param>
        private void CharacterFileUpdatedCallback(object sender, FileSystemEventArgs e)
        {
            ReloadFromFile();
        }

        /// <summary>
        ///Changes worksafe mode - all the lifting is done right in this method.  Allows the settings object to notify monitors when they need to hide from the man.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void WorksafeChangedCallback(object sender, EventArgs e)
        {
            pbCharImage.Visible = !m_settings.WorksafeMode;
            if (m_settings.WorksafeMode)
            {
                tlpInfo.SetColumnSpan(lblSkillHeader, 1);
                tlpInfo.SetColumn(lblSkillHeader, 1);
            }
            else
            {
                tlpInfo.SetColumn(lblSkillHeader, 0);
                tlpInfo.SetColumnSpan(lblSkillHeader, 2);
            }
        }

        /// <summary>
        /// The currently training skill has changed.  Update ineve.
        /// This could probably take over some of the work currently done in CharacterSkillChangedCallback.
        /// </summary>
        /// <param name="sender">The sender - ignored.</param>
        /// <param name="?">The <see cref="System.EventArgs"/> instance containing the event data - ignored.</param>
        private void TrainingSkillChangedCallback(object sender, EventArgs args)
        {
            //will need this if we ever do UI things here, which we probably will.
            /*if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(delegate
                                                  {
                                                      CharacterSkillChangedCallback(sender, e);
                                                  }));
                return;
            }
             */ 
            
            if (m_session != null && m_settings.GetCharacterSettings(m_charName).IneveSync)
                m_session.UpdateIneveAsync(m_grandCharacterInfo);

        }

        /// <summary>
        /// Handles everything when the character's training skill has changed.  Called by GrandCharacterInfo.
        /// </summary>
        /// <remarks>Another high-complexity method for us to look at.</remarks>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EVEMon.Common.SkillChangedEventArgs"/> instance containing the event data.</param>
        private void CharacterSkillChangedCallback(object sender, SkillChangedEventArgs e)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(delegate
                                                  {
                                                      CharacterSkillChangedCallback(sender, e);
                                                  }));
                return;
            }

            EnableButtons();
            lbSkills.BeginUpdate();
            try
            {
                foreach (Skill gs in e.SkillList)
                {
                    SkillGroup gsg = gs.SkillGroup;

                    if (gs.Known)
                    {
                        // Find the existing listbox item... if the group isn't collapsed
                        if (!m_groupCollapsed.ContainsKey(gsg) || m_groupCollapsed[gsg] == false)
                        {
                            int lbIndex = -1;
                            int shouldInsertAt = -1;
                            bool shouldInsertSkillGroup = true;
                            bool inMySkillGroup = false;
                            bool found = false;
                            for (int i = 0; i < lbSkills.Items.Count; i++)
                            {
                                object o = lbSkills.Items[i];
                                if (o == gs)
                                {
                                    shouldInsertSkillGroup = false;
                                    lbIndex = i;
                                    found = true;
                                    break;
                                }
                                else if (o == gsg)
                                {
                                    inMySkillGroup = true;
                                    shouldInsertSkillGroup = false;
                                }
                                else if (o is SkillGroup && ((SkillGroup)o).Name.CompareTo(gsg.Name) > 0)
                                {
                                    shouldInsertAt = i;
                                    shouldInsertSkillGroup = (!inMySkillGroup);
                                    break;
                                }
                                else if (inMySkillGroup && o is Skill &&
                                         ((Skill)o).Name.CompareTo(gs.Name) > 0)
                                {
                                    shouldInsertAt = i;
                                    shouldInsertSkillGroup = false;
                                    break;
                                }
                            }

                            if (shouldInsertSkillGroup)
                            {
                                if (shouldInsertAt >= 0)
                                {
                                    lbSkills.Items.Insert(shouldInsertAt, gsg);
                                    shouldInsertAt++;
                                }
                                else
                                {
                                    lbSkills.Items.Add(gsg);
                                    shouldInsertAt = -1;
                                }
                            }
                            if (!found)
                            {
                                if (shouldInsertAt >= 0)
                                {
                                    lbSkills.Items.Insert(shouldInsertAt, gs);
                                    lbIndex = shouldInsertAt;
                                }
                                else
                                {
                                    lbSkills.Items.Add(gs);
                                    lbIndex = lbSkills.Items.Count - 1;
                                }
                            }
                            lbSkills.Invalidate(lbSkills.GetItemRectangle(lbIndex));
                        }
                    }

                    if (gs.InTraining)
                    {
                        m_skillTrainingName = gs.Name + " " + Skill.GetRomanForInt(gs.TrainingToLevel);
                        lblTrainingSkill.Text = m_skillTrainingName;

                        double spPerHour = 60 * (m_grandCharacterInfo.GetEffectiveAttribute(gs.PrimaryAttribute) +
                                               (m_grandCharacterInfo.GetEffectiveAttribute(gs.SecondaryAttribute) / 2));
                        lblSPPerHour.Text = Convert.ToInt32(Math.Round(spPerHour)).ToString() + " SP/Hour";
                        m_estimatedCompletion = gs.EstimatedCompletion;

                        //downtime finish?
                        DateTime universalFinish = m_estimatedCompletion.ToUniversalTime();
                        //lblScheduleWarning.Visible = (universalFinish.Hour == 11);

                        bool isBlocked = (universalFinish.Hour == 11);
                        for (int i = 0; i < m_settings.Schedule.Count; i++)
                        {
                            ScheduleEntry temp = m_settings.Schedule[i];
                            if (temp.GetType() == typeof(SimpleScheduleEntry))
                            {
                                SimpleScheduleEntry x = (SimpleScheduleEntry)temp;
                                if ((x.ScheduleEntryOptions & ScheduleEntryOptions.Blocking) != 0)
                                {
                                    if ((x.ScheduleEntryOptions & ScheduleEntryOptions.EVETime) != 0)
                                    {
                                        if (x.StartDateTime <= universalFinish && universalFinish <= x.EndDateTime)
                                        {
                                            // This blocks in EVE Time
                                            isBlocked = true;
                                        }
                                    }
                                    else
                                    {
                                        if (x.StartDateTime <= m_estimatedCompletion && m_estimatedCompletion <= x.EndDateTime)
                                        {
                                            // This blocks in local Time
                                            isBlocked = true;
                                        }
                                    }
                                }
                            }
                            else if (temp.GetType() == typeof(RecurringScheduleEntry))
                            {
                                RecurringScheduleEntry x = (RecurringScheduleEntry)temp;
                                if ((x.ScheduleEntryOptions & ScheduleEntryOptions.Blocking) != 0)
                                {
                                    if ((x.ScheduleEntryOptions & ScheduleEntryOptions.EVETime) != 0)
                                    {
                                        // Still needs to be written
                                        // This needs to check to see if universalFinish coincides with
                                        // any of the scheduled periods that this entry represents
                                    }
                                    else
                                    {
                                        // Still needs to be written
                                        // This needs to check to see if m_estimatedCompletion coincides with
                                        // any of the scheduled periods that this entry represents
                                    }
                                }
                            }
                        }
                        lblScheduleWarning.Visible = isBlocked;
                        CalculateLcdData();

                        pnlTraining.Visible = true;
                    }
                    else if (m_grandCharacterInfo.CurrentlyTrainingSkill == null)
                    {
                        m_skillTrainingName = String.Empty;
                        m_estimatedCompletion = DateTime.MaxValue;
                        pnlTraining.Visible = false;
                    }
                }
            }
            finally
            {
                lbSkills.EndUpdate();
            }

            UpdateSkillHeaderStats();
            UpdateCachedCopy();
        }

        /// <summary>
        /// Handle the event when a character's attributes change.  Called by GrandCharacterInfo.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void AttributeChangedCallback(object sender, EventArgs e)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(delegate
                                                  {
                                                      AttributeChangedCallback(sender, e);
                                                  }));
                return;
            }

            EnableButtons();
            SetAttributeLabel(lblIntelligence, EveAttribute.Intelligence);
            SetAttributeLabel(lblCharisma, EveAttribute.Charisma);
            SetAttributeLabel(lblPerception, EveAttribute.Perception);
            SetAttributeLabel(lblMemory, EveAttribute.Memory);
            SetAttributeLabel(lblWillpower, EveAttribute.Willpower);
            UpdateCachedCopy();
        }

        /// <summary>
        /// Handles the character's wallet changing, hopefully for the better.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void BalanceChangedCallback(object sender, EventArgs e)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(delegate
                                                  {
                                                      BalanceChangedCallback(sender, e);
                                                  }));
                return;
            }

            EnableButtons();
            lblBalance.Text = "Balance: " + m_grandCharacterInfo.Balance.ToString("#,##0.00") + " ISK";

            UpdateCachedCopy();
        }

        /// <summary>
        /// Invoked when the character's bio info changes.  Used by GrandCharacterInfo
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void BioInfoChangedCallback(object sender, EventArgs e)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(delegate
                                                  {
                                                      BioInfoChangedCallback(sender, e);
                                                  }));
                return;
            }

            EnableButtons();
            if (m_grandCharacterInfo.IsCached)
            {
                lblCharacterName.Text = m_grandCharacterInfo.Name + " (cached)";
            }
            else
            {
                lblCharacterName.Text = m_grandCharacterInfo.Name;
            }
            lblBioInfo.Text = m_grandCharacterInfo.Gender + " " +
                              m_grandCharacterInfo.Race + " " +
                              m_grandCharacterInfo.Bloodline;
            lblCorpInfo.Text = "Corporation: " + m_grandCharacterInfo.CorporationName;

            UpdateCachedCopy();
        }

        /// <summary>
        /// Update the character image in the event that the remote character image was (probably) retrieved.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="i">The retrieved image.</param>
        private void GotCharacterImage(EveSession sender, Image newImage)
        {
            string cacheFileName = Settings.EveMonData + "\\cache\\" + this.GrandCharacterInfo.CharacterId.ToString() + ".png";

            //the image was not retrieved - go to plan B
            if (newImage == null)
            {
                //Can we get it from the cache?
                if (File.Exists(cacheFileName))
                {
                    newImage = PortraitFromCache(cacheFileName);
                }
                //the image is not in the cache
                if (newImage == null)
                {
                    newImage = pbCharImage.InitialImage;
                    SavePortraitToCache(newImage, cacheFileName);
                }
            }
            else
            {
                SavePortraitToCache(newImage, cacheFileName);
            }
            //whatever happened, show a portrait
            pbCharImage.Image = newImage;
            m_updatingPortrait = false;
        }

        /// <summary>
        /// Open the character portrait from the EVEMon cache
        /// </summary>
        /// <returns>The character portrait as an Image object</returns>
        private static Image PortraitFromCache(string cacheFile)
        {
            Image newImage;
            try
            {
                FileStream fs = new FileStream(cacheFile, FileMode.Open);
                newImage = Image.FromStream(fs, true);
                fs.Close();
                fs.Dispose();
                return newImage;
            }
            catch (Exception e)
            {
                ExceptionHandler.LogException(e, false);
                return null;
            }
        }

        /// <summary>
        /// Save the specified image to the EVEMon cache as this character's portrait
        /// </summary>
        /// <param name="newImage">The new portrait image.</param>
        private static void SavePortraitToCache(Image newImage, string cacheFile)
        {
            try
            {
                //make their character a "!" in the cache
                FileStream fs = new FileStream(cacheFile, FileMode.Create);
                newImage.Save(fs, ImageFormat.Png);
                fs.Close();
                fs.Dispose();
            }
            catch (Exception e)
            {
                ExceptionHandler.LogException(e, false);
            }
        }

        /// <summary>
        /// Callback for the event where a character update failed.  Should be used via Invoke.
        /// </summary>
        private void CharacterDownloadFailedCallback()
        {
            ttToolTip.SetToolTip(pbThrobber, "Could not get character data!\nClick to try again.");
            ttToolTip.IsBalloon = true;
            tmrUpdateCharacter.Interval = 1000 * 60 * 30;
            tmrUpdateCharacter.Enabled = true;
            SetErrorThrobber();
            this.m_grandCharacterInfo.check_old_skill();
        }
        #endregion

        #region Control/Component Event Handlers

        /// <summary>
        /// Handles animation for the throbber control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void tmrThrobber_Tick(object sender, EventArgs e)
        {
            if (m_currentlyVisible)
            {
                tmrThrobber.Enabled = false;
                if (m_throbberRunning)
                {
                    m_throbberFrame = ((m_throbberFrame + 1) % 8);
                    pbThrobber.Image = ThrobberImages[m_throbberFrame + 1];
                    tmrThrobber.Enabled = true;
                }
                else if (m_throbberError)
                {
                    bool blinkState = (DateTime.Now.Millisecond > 500);
                    if (!blinkState)
                    {
                        pbThrobber.Image = null;
                    }
                    else
                    {
                        pbThrobber.Image = ThrobberImages[0];
                    }
                    tmrThrobber.Enabled = true;
                }
                else
                {
                    pbThrobber.Image = ThrobberImages[0];
                    m_throbberFrame = 0;
                }
            }
        }

        /// <summary>
        /// Handles the Click event of the pbThrobber control.  If we're in an OK state, hit eve-o, otherwise, show the throbber menu.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void pbThrobber_Click(object sender, EventArgs e)
        {
            if (!m_throbberError)
            {
                tmrUpdate_Tick(null, null);
                UpdateThrobberLabel();
            }
            else
            {
                cmsThrobberMenu.Show(MousePosition);
            }
        }

        private void lbSkills_MouseEnter(object sender, EventArgs e)
        {
            ttToolTip.Active = false;
        }

        private void lbSkills_MouseLeave(object sender, EventArgs e)
        {
            ttToolTip.Active = false;
        }

        private void pbThrobber_MouseEnter(object sender, EventArgs e)
        {
            ttToolTip.IsBalloon = true;
            ttToolTip.Active = true;
        }

        private void miHitEveO_Click(object sender, EventArgs e)
        {
            tmrUpdate_Tick(this, new EventArgs());
        }

        /// <summary>
        /// Handles the Click event of the miChangeInfo control.  Displays a new login window and gets username/password from the user.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void miChangeInfo_Click(object sender, EventArgs e)
        {
            using (ChangeLoginWindow f = new ChangeLoginWindow())
            {
                f.CharacterName = m_grandCharacterInfo.Name;
                DialogResult dr = f.ShowDialog();
                if (dr == DialogResult.OK)
                {
                    m_cli.Username = f.Username;
                    m_cli.Password = f.Password;
                    m_session = EveSession.GetSession(f.Username, f.Password);
                    m_charId = -1;
                    tmrUpdate_Tick(this, new EventArgs());

                    m_settings.Save();
                }
            }
        }

        /// <summary>
        /// Handles the Click event of the miManualImplants control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void miManualImplants_Click(object sender, EventArgs e)
        {
            using (ManualImplantWindow f = new ManualImplantWindow(m_grandCharacterInfo.AttributeBonuses))
            {
                DialogResult dr = f.ShowDialog();
                if (dr == DialogResult.OK)
                {
                    m_grandCharacterInfo.SuppressEvents();
                    try
                    {
                        for (int i = 0; i < m_grandCharacterInfo.AttributeBonuses.Count; i++)
                        {
                            GrandEveAttributeBonus geab = m_grandCharacterInfo.AttributeBonuses[i];
                            if (geab.Manual)
                            {
                                m_grandCharacterInfo.AttributeBonuses.RemoveAt(i);
                                i--;
                            }
                        }
                        foreach (GrandEveAttributeBonus b in f.ResultBonuses)
                        {
                            if (b.Manual)
                                m_grandCharacterInfo.AttributeBonuses.Add(b);
                        }
                    }
                    finally
                    {
                        m_grandCharacterInfo.ResumeEvents();
                    }
                }
            }
        }

        /// <summary>
        /// Handles the DrawItem event of the lbSkills control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.DrawItemEventArgs"/> instance containing the event data.</param>
        private void lbSkills_DrawItem(object sender, DrawItemEventArgs e)
        {
            if (e.Index < 0)
            {
                return;
            }
            object item = lbSkills.Items[e.Index];

            if (item is SkillGroup)
            {
                ((SkillGroup)item).Draw(e);
            }
            else if (item is Skill)
            {
                ((Skill)item).Draw(e);
            }
        }

        /// <summary>
        /// Handles the MeasureItem event of the lbSkills control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.MeasureItemEventArgs"/> instance containing the event data.</param>
        private void lbSkills_MeasureItem(object sender, MeasureItemEventArgs e)
        {
            if (e.Index < 0)
            {
                return;
            }
            object item = lbSkills.Items[e.Index];
            if (item is SkillGroup)
            {
                e.ItemHeight = SkillGroup.Height;
            }
            else if (item is Skill)
            {
                e.ItemHeight = Skill.Height;
            }
        }

        /// <summary>
        /// Handles the MouseWheel event of the lbSkills control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.MouseEventArgs"/> instance containing the event data.</param>
        private void lbSkills_MouseWheel(object sender, MouseEventArgs e)
        {
            // Update the drawing based upon the mouse wheel scrolling.
            int numberOfItemLinesToMove = e.Delta * SystemInformation.MouseWheelScrollLines / 120;
            int Lines = numberOfItemLinesToMove;
            int direction = numberOfItemLinesToMove / Math.Abs(numberOfItemLinesToMove);
            int[] numberOfPixelsToMove = new int[numberOfItemLinesToMove * direction];
            for (int i = 1; i <= Math.Abs(numberOfItemLinesToMove); i++)
            {
                object item = null;
                if (direction == Math.Abs(direction))
                {
                    // Going up
                    if (lbSkills.TopIndex - i >= 0)
                    {
                        item = lbSkills.Items[lbSkills.TopIndex - i];
                    }
                }
                else
                {
                    // Going down
                    int h = 0; // height of items from current topindex inclusive
                    for (int j = lbSkills.TopIndex + i - 1; j < lbSkills.Items.Count; j++)
                    {
                        if (lbSkills.Items[j] is SkillGroup)
                        {
                            h += SkillGroup.Height;
                        }
                        else if (lbSkills.Items[j] is Skill)
                        {
                            h += Skill.Height;
                        }
                    }
                    if (h > lbSkills.ClientSize.Height)
                    {
                        item = lbSkills.Items[lbSkills.TopIndex + i - 1];
                    }
                }
                if (item != null)
                {
                    if (item is SkillGroup)
                    {
                        numberOfPixelsToMove[i - 1] = SkillGroup.Height * direction;
                    }
                    else if (item is Skill)
                    {
                        numberOfPixelsToMove[i - 1] = Skill.Height * direction;
                    }
                }
                else
                {
                    Lines -= direction;
                }
            }
            if (Lines != 0)
            {
                // The Array 'numberOfPixelsToMove' contains the number of pixels the
                // list box 'lbSkills' needs to scroll for each line... the question
                // is, how to tell it that and get it to do so smoothly.
                /* // This doesn't work...
                for (int i = 0; i < Math.Abs(Lines); i++)
                {
                    System.Drawing.Drawing2D.Matrix translateMatrix = new System.Drawing.Drawing2D.Matrix();
                    translateMatrix.Translate(0, numberOfPixelsToMove[i]);
                    mousePath.Transform(translateMatrix);
                } */
                lbSkills.Invalidate();
                // invalidate is a temporary fix that does give limited functionality for purpose.
            }
        }

        /// <summary>
        /// Handles the Click event of the btnPlan control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void btnPlan_Click(object sender, EventArgs e)
        {
            Plan p = null;
            using (PlanSelectWindow psw = new PlanSelectWindow(m_settings, m_grandCharacterInfo))
            {
                if (m_cfi != null)
                {
                    psw.CharKey = m_cfi.Filename;
                }
                DialogResult dr = psw.ShowDialog();
                if (dr == DialogResult.Cancel)
                {
                    return;
                }
                p = psw.ResultPlan;
            }
            if (p == null || p.Name == null)
            {
                bool doAgain = true;
                while (doAgain)
                {
                    using (NewPlanWindow npw = new NewPlanWindow())
                    {
                        DialogResult dr = npw.ShowDialog();
                        if (dr == DialogResult.Cancel)
                        {
                            return;
                        }
                        string planName = npw.Result;

                        if (p == null)
                        {
                            p = new Plan();
                        }
                        try
                        {
                            if (m_cfi == null)
                            {
                                m_settings.AddPlanFor(m_grandCharacterInfo.Name, p, planName);
                            }
                            else
                            {
                                m_settings.AddPlanFor(m_cfi.Filename, p, planName);
                            }
                            doAgain = false;
                        }
                        catch (ApplicationException err)
                        {
                            ExceptionHandler.LogException(err, true);
                            DialogResult xdr =
                                MessageBox.Show(err.Message, "Failed to Add Plan", MessageBoxButtons.OKCancel,
                                                MessageBoxIcon.Error);
                            if (xdr == DialogResult.Cancel)
                            {
                                return;
                            }
                        }
                    }
                }
            }

            p.ShowEditor(m_settings, m_grandCharacterInfo);
        }

        /// <summary>
        /// Handles the ContextMenuShowing event of the btnPlan control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void btnPlan_ContextMenuShowing(object sender, EventArgs e)
        {
            ContextMenuStrip plans = new ContextMenuStrip();

            foreach (string plan in m_settings.GetPlansForCharacter(m_grandCharacterInfo.Name))
            {

                ToolStripMenuItem planItem = new ToolStripMenuItem(plan);
                planItem.Click += delegate(object o, EventArgs ev)
                {
                    ToolStripMenuItem item = o as ToolStripMenuItem;
                    m_settings.GetPlanByName(m_grandCharacterInfo.Name, item.Text).ShowEditor(m_settings, m_grandCharacterInfo);
                };
                plans.Items.Add(planItem);
            }

            btnPlan.ContextMenuStrip = plans;
        }


        private void btnMoreOptions_Click(object sender, EventArgs e)
        {
            cmsMoreOptions.Show(btnMoreOptions,
                                btnMoreOptions.PointToClient(MousePosition), ToolStripDropDownDirection.Default);
        }


        private void lbSkills_MouseDown(object sender, MouseEventArgs e)
        {
            lbSkills_MouseClick(sender, e);
        }

        /// <summary>
        /// Handles the MouseClick event of the lbSkills control.
        /// </summary>
        /// <remarks>TODO:  There are opportunities for improvement here.</remarks>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Forms.MouseEventArgs"/> instance containing the event data.</param>
        private void lbSkills_MouseClick(object sender, MouseEventArgs e)
        {
            int index = lbSkills.IndexFromPoint(e.X, e.Y);
            object item;
            if (index < 0 || index >= lbSkills.Items.Count)
            {
                item = null;
            }
            else
            {
                item = lbSkills.Items[index];
            }

            ttToolTip.IsBalloon = true;
            ttToolTip.UseAnimation = true;
            ttToolTip.UseFading = true;
            ttToolTip.AutoPopDelay = 10000;

            if (item is SkillGroup)
            {
                SkillGroup sg = (SkillGroup)item;
                if (e.Button == MouseButtons.Left)
                {
                    ToggleGroupExpandCollapse(sg);
                    return;
                }
                else if (e.Button == MouseButtons.Right)
                {
                    Rectangle itemRect = lbSkills.GetItemRectangle(lbSkills.Items.IndexOf(item));
                    Rectangle buttonRect = sg.GetButtonRectangle(itemRect);
                    if (buttonRect.Contains(e.Location))
                    {
                        ToggleGroupExpandCollapse(sg);
                        return;
                    }

                    int TotalPoints = 0;
                    double percentDonePoints = 0.0;
                    double percentDoneSkills = 0.0;

                    foreach (Skill s in sg)
                    {
                        TotalPoints += s.GetPointsRequiredForLevel(5);
                    }

                    if (sg.GetTotalPoints() < TotalPoints)
                    {
                        percentDonePoints = (Convert.ToDouble(sg.GetTotalPoints()) / Convert.ToDouble(TotalPoints) * 100);
                        percentDoneSkills = Convert.ToDouble(sg.KnownCount) / Convert.ToDouble(sg.Count);

                        string SkillGroupStats =
                            String.Format("Points Completed: {0}/{1} ({2}%)\nSkills Known: {3}/{4} ({5})",
                                          sg.GetTotalPoints().ToString("#,##0"), TotalPoints.ToString("#,##0"),
                                          percentDonePoints.ToString("N3"), sg.KnownCount.ToString("#"),
                                          sg.Count.ToString("#"), percentDoneSkills.ToString("P0"));

                        ttToolTip.SetToolTip(lbSkills, SkillGroupStats);
                        ttToolTip.Active = true;
                    }
                    else // we must have learned all the skills in this group to level 5
                    {
                        //I wish I could test this :)
                        // - save xml, edit to make all skills in a group l5, import?
                        string Done = String.Format("Skill Group completed: {0}/{1} (100%)\nSkills: {2}/{3} (100%)",
                                                    sg.GetTotalPoints().ToString("#,##0"), TotalPoints.ToString("#,##0"),
                                                    sg.KnownCount.ToString("#"), sg.Count.ToString("#"));
                        ttToolTip.Active = true;
                        ttToolTip.SetToolTip(lbSkills, Done);
                    }
                }
            }
            else if (item is Skill)
            {
                Skill s = (Skill)item;
                double percentDone = 0.0;
                int NextLevel = 0;
                int CurrentSP = s.CurrentSkillPoints;
                int reqToThisLevel = s.GetPointsRequiredForLevel(s.Level);
                int PointsRemain = 0;
                double spPerHour = 60 * (m_grandCharacterInfo.GetEffectiveAttribute(s.PrimaryAttribute) +
                                       (m_grandCharacterInfo.GetEffectiveAttribute(s.SecondaryAttribute) / 2));
                string SPPerHour = " (" + Convert.ToInt32(Math.Round(spPerHour)).ToString() + " SP/Hour)";

                if (CurrentSP > s.GetPointsRequiredForLevel(s.Level))
                {
                    //We must have completed some, but not all, of level II, III or IV
                    NextLevel = s.Level + 1;

                    percentDone = s.GetPercentDone();
                    PointsRemain = s.GetPointsRequiredForLevel(NextLevel) - s.CurrentSkillPoints;
                    string CurrentlyDone =
                        String.Format("Partially Completed lvl {0}: {1}/{2} ({3})",
                                      Skill.GetRomanForInt(NextLevel),
                                      s.CurrentSkillPoints.ToString("#,##0"),
                                      s.GetPointsRequiredForLevel(NextLevel).ToString("#,##0"),
                                      percentDone.ToString("P0"));
                    string ToNextLevel =
                        String.Format("To Level {0}: {1} Skill Points remaining",
                                      Skill.GetRomanForInt(NextLevel), PointsRemain.ToString("#,##0"));
                    ttToolTip.Active = true;
                    ttToolTip.SetToolTip(lbSkills,
                                         CurrentlyDone + "\n" + ToNextLevel + "\nTraining Time remaining: " +
                                         Skill.TimeSpanToDescriptiveText(
                                             s.GetTrainingTimeToLevel(NextLevel),
                                             DescriptiveTextOptions.IncludeCommas |
                                             DescriptiveTextOptions.UppercaseText) + "\n" +
                                         s.DescriptionNl.ToString() + "\nPrimary: " +
                                         s.PrimaryAttribute.ToString() + ", Secondary: " +
                                         s.SecondaryAttribute.ToString() + SPPerHour);
                }
                else if (CurrentSP == s.GetPointsRequiredForLevel(s.Level))
                // We've completed all the skill points for the current level
                {
                    if (s.Level != 5)
                    {
                        NextLevel = s.Level + 1;
                        percentDone = s.GetPercentDone();
                        PointsRemain = s.GetPointsRequiredForLevel(NextLevel) - s.CurrentSkillPoints;
                        string CurrentlyDone =
                            String.Format("Completed lvl {0}: {1}/{2} ({3})",
                                          Skill.GetRomanForInt(s.Level),
                                          s.CurrentSkillPoints.ToString("#,##0"),
                                          s.GetPointsRequiredForLevel(s.Level).ToString("#,##0"),
                                          percentDone.ToString("P0"));
                        string ToNextLevel =
                            String.Format("To Level {0}: {1} Skill Points required",
                                          Skill.GetRomanForInt(NextLevel),
                                          PointsRemain.ToString("#,##0"));
                        ttToolTip.Active = true;
                        ttToolTip.SetToolTip(lbSkills,
                                             CurrentlyDone + "\n" + ToNextLevel + "\nTraining Time: " +
                                             Skill.TimeSpanToDescriptiveText(
                                                 s.GetTrainingTimeToLevel(NextLevel),
                                                 DescriptiveTextOptions.IncludeCommas |
                                                 DescriptiveTextOptions.UppercaseText) + "\n" +
                                             s.DescriptionNl.ToString() + "\nPrimary: " +
                                             s.PrimaryAttribute.ToString() + ", Secondary: " +
                                             s.SecondaryAttribute.ToString() + SPPerHour);
                    }
                    else // training completed
                    {
                        ttToolTip.Active = true;
                        ttToolTip.SetToolTip(lbSkills,
                                             String.Format(
                                                 "Level V Complete: {0}/{1} (100%)\nNo further training required\n{2}\nPrimary: {3}, Secondary: {4} {5}",
                                                 s.CurrentSkillPoints.ToString("#,##0"),
                                                 s.GetPointsRequiredForLevel(5).ToString("#,##0"),
                                                 s.DescriptionNl.ToString(), s.PrimaryAttribute.ToString(),
                                                 s.SecondaryAttribute.ToString(), SPPerHour));
                    }
                }
                else // training hasn't got past level 1 yet
                {
                    NextLevel = s.Level + 1; //this should always be 1

                    percentDone = Convert.ToDouble(CurrentSP) /
                                  Convert.ToDouble(s.GetPointsRequiredForLevel(NextLevel));
                    PointsRemain = s.GetPointsRequiredForLevel(NextLevel) - s.CurrentSkillPoints;
                    string CurrentlyDone =
                        String.Format("Partially Completed lvl {0}: {1}/{2} ({3})",
                                      Skill.GetRomanForInt(NextLevel),
                                      s.CurrentSkillPoints.ToString("#,##0"),
                                      s.GetPointsRequiredForLevel(NextLevel).ToString("#,##0"),
                                      percentDone.ToString("P0"));
                    string ToNextLevel =
                        String.Format("To Level {0}: {1} Skill Points remaining",
                                      Skill.GetRomanForInt(NextLevel), PointsRemain.ToString("#,##0"));
                    ttToolTip.Active = true;
                    ttToolTip.SetToolTip(lbSkills,
                                         CurrentlyDone + "\n" + ToNextLevel + "\nTraining Time remaining: " +
                                         Skill.TimeSpanToDescriptiveText(
                                             s.GetTrainingTimeToLevel(NextLevel),
                                             DescriptiveTextOptions.IncludeCommas |
                                             DescriptiveTextOptions.UppercaseText) + "\n" +
                                         s.DescriptionNl.ToString() + "\nPrimary: " +
                                         s.PrimaryAttribute.ToString() + ", Secondary: " +
                                         s.SecondaryAttribute.ToString() + SPPerHour);
                }
            }
        }


        private void miUpdatePicture_Click(object sender, EventArgs e)
        {
            UpdateCharacterImageRemote();
        }


        private void miUpdatePictureFromEVECache_Click(object sender, EventArgs e)
        {
            UpdateCharacterImageLocal();
        }


        private void miSetEVEFolder_Click(object sender, EventArgs e)
        {
            RequestEVEFolder();
        }


        private void pbCharImage_Click(object sender, EventArgs e)
        {
            cmsPictureOptions.Show(MousePosition);
        }

        /// <summary>
        /// Handles the MouseHover event of the lblAttribute control.  Pops up a little tooltip containing everything the person could possibly want to know
        /// about their character's attributes.
        /// </summary>
        /// <remarks>Recalculates everything every time the label pops - so it's inefficient, but responsive enough for this purpose.</remarks>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void lblAttribute_MouseHover(object sender, EventArgs e)
        {
            Label lblAttrib = (Label)sender;
            EveAttribute eveAttribute = EveAttribute.None;
            switch (lblAttrib.Name)
            {
                case "lblIntelligence": eveAttribute = EveAttribute.Intelligence; break;
                case "lblCharisma": eveAttribute = EveAttribute.Charisma; break;
                case "lblMemory": eveAttribute = EveAttribute.Memory; break;
                case "lblWillpower": eveAttribute = EveAttribute.Willpower; break;
                case "lblPerception": eveAttribute = EveAttribute.Perception; break;
                default: break;
            }

            StringBuilder sb = new StringBuilder();

            sb.Append(eveAttribute.ToString());

            sb.Append(": ");
            sb.Append(m_grandCharacterInfo.GetEffectiveAttribute(eveAttribute).ToString("0.00"));

            int baseAtt = m_grandCharacterInfo.GetBaseAttribute(eveAttribute);
            double fromSkills = m_grandCharacterInfo.GetEffectiveAttribute(eveAttribute, null, false, false) - baseAtt;
            double learning = m_grandCharacterInfo.LearningBonus;
            double implant = m_grandCharacterInfo.getImplantValue(eveAttribute);

            sb.Append(" [");
            if (learning > 0.0) sb.Append("(");
            sb.Append(baseAtt.ToString("0"));
            sb.Append(" base + ");
            sb.Append(fromSkills.ToString("0"));
            sb.Append(" skills + ");
            sb.Append(implant.ToString("0"));
            if (learning > 0.0)
            {
                sb.Append(" implants) * ");
                sb.Append(learning.ToString("0.00"));
                sb.Append(" from learning bonus");
            }
            sb.Append("]");

            ttToolTip.SetToolTip(lblAttrib, sb.ToString());
            ttToolTip.IsBalloon = false;
            ttToolTip.Active = true;

        }


        private void tsbIneveSync_CheckedChanged(object sender, EventArgs e)
        {
            m_settings.GetCharacterSettings(m_charName).IneveSync = tsbIneveSync.Checked;
        }


        private void btnSave_Click(object sender, EventArgs e)
        {
            sfdSaveDialog.FileName = m_grandCharacterInfo.Name;
            sfdSaveDialog.FilterIndex = (int)SaveFormat.Xml;
            DialogResult dr = sfdSaveDialog.ShowDialog();
            if (dr == DialogResult.OK)
            {
                SaveFile((SaveFormat)sfdSaveDialog.FilterIndex, sfdSaveDialog.FileName);
            }
        }

        /// <summary>
        /// Handles the Tick event of the tmrUpdate control.  This is the one that updates the character itself.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void tmrUpdate_Tick(object sender, EventArgs e)
        {
            if (m_throbberRunning)
            {
                return;
            }

            tmrUpdateCharacter.Enabled = false;
            StartThrobber();
            if (m_charId < 0)
            {
                DownloadCharacter();
            }
            else
            {
                UpdateGrandCharacterInfo();
            }
        }

        /// <summary>
        /// Things that happen every second:
        /// Update LCD Data, Update time remaining and throbber labels, check your skills, so on, so forth
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void tmrTick_Tick(object sender, EventArgs e)
        {
            CalculateLcdData();

            if (m_currentlyVisible)
            {
                UpdateTimeRemainingLabels();
                UpdateThrobberLabel();

                Skill trainingSkill = m_grandCharacterInfo.CurrentlyTrainingSkill;
                if (trainingSkill != null)
                {
                    if (trainingSkill.CurrentSkillPoints != m_lastTickSPPaint)
                    {
                        m_lastTickSPPaint = trainingSkill.CurrentSkillPoints;
                        int idx = lbSkills.Items.IndexOf(trainingSkill);
                        if (idx >= 0)
                        {
                            lbSkills.Invalidate(lbSkills.GetItemRectangle(idx));
                        }
                        int sgidx = lbSkills.Items.IndexOf(trainingSkill.SkillGroup);
                        lbSkills.Invalidate(lbSkills.GetItemRectangle(sgidx));
                        UpdateSkillHeaderStats();
                    }
                }
            }
            if (m_grandCharacterInfo.DL_Complete && m_estimatedCompletion < DateTime.Now &&
                m_grandCharacterInfo.CurrentlyTrainingSkill != null)
            {
                // Trigger event on skill completion
                // The 'event' for a skill completion is in GrandCharacterInfo.cs, depending on what is wanted
                // it should be triggered from there as all skill manipulation is done in that file.

                // The following line triggers the required code in GrandCharacterInfo.cs
                m_grandCharacterInfo.trigger_skill_complete(m_charName, m_grandCharacterInfo.CurrentlyTrainingSkill.Name);

                UpdateSkillHeaderStats();
            }
            if (m_updatingPortrait == false && pbCharImage.Image == null)
            {
                m_updatingPortrait = true;
                GetCharacterImage();
            }
            if (Program.LCD != null)
            {
                Program.LCD.Update();
            }
        }
        #endregion

        #region Simple Properties
        /// <summary>
        /// Gets the grand character info.
        /// </summary>
        /// <value>The grand character info.</value>
        public CharacterInfo GrandCharacterInfo
        {
            get { return m_grandCharacterInfo; }
        }

        public TimeSpan ShortTimeSpan { get { return m_lcdTimeSpan; } }
        #endregion

        #region Saving the Character

        /// <summary>
        /// Saves the character skill list as a PNG screenshot
        /// </summary>
        /// <param name="fileName">Name of the file.</param>
        private void SavePNGScreenshot(string fileName)
        {
            int cachedHeight = lbSkills.Height;
            lbSkills.Dock = System.Windows.Forms.DockStyle.None;
            lbSkills.Height = lbSkills.PreferredHeight;
            lbSkills.Update();

            Bitmap bitmap = new Bitmap(lbSkills.Width, lbSkills.PreferredHeight);
            lbSkills.DrawToBitmap(bitmap, new Rectangle(0, 0, lbSkills.Width, lbSkills.PreferredHeight));
            bitmap.Save(fileName, System.Drawing.Imaging.ImageFormat.Png);

            lbSkills.Dock = System.Windows.Forms.DockStyle.Fill;
            lbSkills.Height = cachedHeight;
            lbSkills.Update();

            this.Invalidate();
        }

        /// <summary>
        /// Saves character data as a text file
        /// </summary>
        /// <param name="fileName">Name of the file.</param>
        private void SaveTextFile(string fileName)
        {
            SerializableCharacterInfo ci = m_grandCharacterInfo.ExportSerializableCharacterInfo();
            using (StreamWriter sw = new StreamWriter(fileName, false))
            {
                MethodInvoker writeSep = new MethodInvoker(delegate
                                                               {
                                                                   sw.WriteLine(
                                                                       "=======================================================================");
                                                               });
                MethodInvoker writeSubSep = new MethodInvoker(delegate
                                                                  {
                                                                      sw.WriteLine(
                                                                          "-----------------------------------------------------------------------");
                                                                  });
                sw.WriteLine("BASIC INFO");
                writeSep();
                sw.WriteLine("     Name: {0}", ci.Name);
                sw.WriteLine("   Gender: {0}", ci.Gender);
                sw.WriteLine("     Race: {0}", ci.Race);
                sw.WriteLine("Bloodline: {0}", ci.BloodLine);
                sw.WriteLine("  Balance: {0} ISK", ci.Balance.ToString("#,##0.00"));
                sw.WriteLine();
                sw.WriteLine("Intelligence: {0}", ci.Attributes.AdjustedIntelligence.ToString("#0.00").PadLeft(5));
                sw.WriteLine("    Charisma: {0}", ci.Attributes.AdjustedCharisma.ToString("#0.00").PadLeft(5));
                sw.WriteLine("  Perception: {0}", ci.Attributes.AdjustedPerception.ToString("#0.00").PadLeft(5));
                sw.WriteLine("      Memory: {0}", ci.Attributes.AdjustedMemory.ToString("#0.00").PadLeft(5));
                sw.WriteLine("   Willpower: {0}", ci.Attributes.AdjustedWillpower.ToString("#0.00").PadLeft(5));
                sw.WriteLine();
                if (ci.AttributeBonuses.Bonuses.Count > 0)
                {
                    sw.WriteLine("IMPLANTS");
                    writeSep();
                    foreach (SerializableEveAttributeBonus tb in ci.AttributeBonuses.Bonuses)
                    {
                        sw.WriteLine("+{0} {1} : {2}", tb.Amount, tb.EveAttribute.ToString().PadRight(13), tb.Name);
                    }
                    sw.WriteLine();
                }
                sw.WriteLine("SKILLS");
                writeSep();
                foreach (SerializableSkillGroup sg in ci.SkillGroups)
                {
                    sw.WriteLine("{0}, {1} Skill{2}, {3} Points",
                                 sg.Name, sg.Skills.Count, sg.Skills.Count > 1 ? "s" : "",
                                 sg.GetTotalPoints().ToString("#,##0"));
                    foreach (SerializableSkill s in sg.Skills)
                    {
                        string skillDesc = s.Name + " " + Skill.GetRomanForInt(s.Level) + " (" +
                                           s.Rank.ToString() + ")";
                        sw.WriteLine(": {0} {1}/{2} Points",
                                     skillDesc.PadRight(40), s.SkillPoints.ToString("#,##0"),
                                     s.SkillLevel5.ToString("#,##0"));
                        if (ci.SkillInTraining != null && ci.SkillInTraining.SkillName == s.Name)
                        {
                            sw.WriteLine(":  (Currently training to level {0}, completes {1})",
                                         Skill.GetRomanForInt(ci.SkillInTraining.TrainingToLevel),
                                         ci.SkillInTraining.EstimatedCompletion.ToString());
                        }
                    }
                    writeSubSep();
                }
            }
        }

        /// <summary>
        /// Saves character to a file in one of a variety of formats.  Does clever things with xsl to avoid having a seperate xml and html method.
        /// </summary>
        /// <param name="saveFormat">The save format.</param>
        /// <param name="fileName">Name of the file.</param>
        private void SaveFile(SaveFormat saveFormat, string fileName)
        {
            if (saveFormat == SaveFormat.Text)
            {
                SaveTextFile(fileName);
                return;
            }
            else if (saveFormat == SaveFormat.Png)
            {
                SavePNGScreenshot(fileName);
                return;
            }
            try
            {
                Stream outerStream;
                XPathDocument xpdoc;
                if (saveFormat == SaveFormat.Xml)
                {
                    outerStream = new FileStream(fileName, FileMode.Create, FileAccess.Write, FileShare.None);
                }
                else
                {
                    outerStream = new MemoryStream(32767);
                }
                try
                {
                    using (XmlTextWriter xtw = new XmlTextWriter(outerStream, Encoding.UTF8))
                    {
                        if (saveFormat == SaveFormat.Xml)
                        {
                            xtw.Indentation = 1;
                            xtw.IndentChar = '\t';
                            xtw.Formatting = Formatting.Indented;
                        }
                        XmlSerializer xs = new XmlSerializer(typeof(SerializableCharacterInfo));
                        XmlSerializerNamespaces ns = new XmlSerializerNamespaces();
                        ns.Add("", "");
                        xs.Serialize(xtw, m_grandCharacterInfo.ExportSerializableCharacterInfo(), ns);
                        xtw.Flush();

                        if (saveFormat == SaveFormat.Xml)
                        {
                            return;
                        }

                        MemoryStream ms = (MemoryStream)outerStream;
                        ms.Position = 0;
                        using (StreamReader tr = new StreamReader(ms))
                        {
                            xpdoc = new XPathDocument(tr);
                        }
                    }
                }
                finally
                {
                    outerStream.Dispose();
                }

                XslCompiledTransform xstDoc2 = new XslCompiledTransform();
                using (
                    Stream s =
                        Assembly.GetExecutingAssembly().GetManifestResourceStream("EVEMon.output-" +
                                                                                  saveFormat.ToString().ToLower() +
                                                                                  ".xsl"))
                using (XmlTextReader xtr = new XmlTextReader(s))
                {
                    xstDoc2.Load(xtr);
                }

                using (StreamWriter sw = new StreamWriter(fileName, false))
                using (XmlTextWriter xtw = new XmlTextWriter(sw))
                {
                    xtw.Indentation = 1;
                    xtw.IndentChar = '\t';
                    xtw.Formatting = Formatting.Indented;
                    xstDoc2.Transform(xpdoc, null, xtw);
                }
            }
            catch (Exception ex)
            {
                ExceptionHandler.LogException(ex, true);
                MessageBox.Show("Failed to save:\n" + ex.Message, "Could not save", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #endregion

        #region Throbber Management
        private bool m_throbberRunning = false;
        private bool m_throbberError = true;
        private int m_throbberFrame = 0;

        private void StartThrobber()
        {
            m_throbberRunning = true;
            tmrThrobber.Enabled = true;
            ttToolTip.SetToolTip(pbThrobber, "Retrieving data from EVE Online...");
            ttToolTip.IsBalloon = true;
        }

        private void StopThrobber()
        {
            m_throbberRunning = false;
            m_throbberError = false;
        }

        private void SetErrorThrobber()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(SetErrorThrobber));
                return;
            }
            m_throbberRunning = false;
            m_throbberError = true;
        }

        private static Image[] m_throbberImages = null;

        public static Image[] ThrobberImages
        {
            get
            {
                if (m_throbberImages == null)
                {
                    InitializeThrobberImages();
                }
                return m_throbberImages;
            }
        }

        private const int THROBBERIMG_WIDTH = 24;
        private const int THROBBERIMG_HEIGHT = 24;

        /// <summary>
        /// Extract the throbber images from the assembly and load them into m_throbberImages
        /// </summary>
        private static void InitializeThrobberImages()
        {
            Assembly asm = Assembly.GetExecutingAssembly();
            using (Stream s = asm.GetManifestResourceStream("EVEMon.throbber.png"))
            using (Image b = Image.FromStream(s, true, true))
            {
                m_throbberImages = new Image[9];
                for (int i = 0; i < 9; i++)
                {
                    Bitmap ib = new Bitmap(THROBBERIMG_WIDTH, THROBBERIMG_HEIGHT);
                    using (Graphics g = Graphics.FromImage(ib))
                    {
                        g.DrawImage(b, new Rectangle(0, 0, THROBBERIMG_WIDTH, THROBBERIMG_HEIGHT),
                                    new Rectangle(i * THROBBERIMG_WIDTH, 0, THROBBERIMG_WIDTH, THROBBERIMG_HEIGHT),
                                    GraphicsUnit.Pixel);
                    }
                    m_throbberImages[i] = ib;
                }
            }
        }
        #endregion

        #region Garbage?

        /*
        public string LcdNote
        {
            get { return m_lcdNote; }
        }

        public TimeSpan LcdTimeSpan
        {
            get { return m_lcdTimeSpan; }
        }
         * */
        #endregion
    }
}
